from octopus.platforms.ETH.explorer import EthereumInfuraExplorer

import unittest


class EthereumExplorerTestCase(unittest.TestCase):

    # please only used this key for octopus - Infura registration is FREE
    explorer = EthereumInfuraExplorer("bHuaQhX91nkQBac8Wtgj")

    '''TODO'''
    block_number = 5100196
    block_hash = '0x98a548cbd0cd385f46c9bf28c16bc36dc6ec27207617e236f527716e617ae91b'
    contract_address = '0xBB9bc244D798123fDe783fCc1C72d3Bb8C189413'
    address = "0x956b6B7454884b734B29A8115F045a95179ea00C"
    tx_hash = '0x345303843c2f3041d12f0c5e6075fd294c2e2ca8cd9b4a9addca3f8caf4380ff'

    def testRPCCommand(self):

        #######################
        # HIGHT-LEVEL METHODS #
        #######################

        self.assertEqual(self.explorer.get_transaction(self.tx_hash), self.explorer.eth_getTransactionByHash(self.tx_hash))
        self.assertEqual(len(self.explorer.get_block_by_hash(self.block_hash)), 20)
        self.assertEqual(len(self.explorer.get_block_by_number(self.block_number)), 20)

        ####################
        # JSON-RPC METHODS #
        ####################

        self.assertEqual(type(self.explorer.web3_clientVersion()), str)
        self.assertEqual(type(self.explorer.web3_sha3('0x' + b'hello world'.hex())), str)
        self.assertEqual(type(self.explorer.net_version()), str)
        self.assertEqual(type(self.explorer.net_listening()), bool)
        self.assertEqual(type(self.explorer.net_peerCount()), int)
        self.assertEqual(type(self.explorer.eth_protocolVersion()), str)
        self.assertEqual(type(self.explorer.eth_syncing()), bool)
        self.assertEqual(type(self.explorer.eth_mining()), bool)
        self.assertEqual(type(self.explorer.eth_hashrate()), int)
        self.assertEqual(type(self.explorer.eth_gasPrice()), int)
        self.assertEqual(type(self.explorer.eth_accounts()), list)
        self.assertEqual(type(self.explorer.eth_blockNumber()), int)
        self.assertEqual(type(self.explorer.eth_getBalance(self.address)), int)
        self.assertEqual(type(self.explorer.eth_getStorageAt("0x295a70b2de5e3953354a6a8344e616ed314d7251", 0, "latest")), str)
        self.assertEqual(type(self.explorer.eth_getTransactionCount(self.address)), int)

        self.assertEqual(self.explorer.eth_getBlockTransactionCountByHash(self.block_hash), 69)
        self.assertEqual(self.explorer.eth_getBlockTransactionCountByNumber(self.block_number), 69)
        self.assertEqual(self.explorer.eth_getUncleCountByBlockHash(self.block_hash), 0)
        self.assertEqual(self.explorer.eth_getUncleCountByBlockNumber(self.block_number), 0)

        self.assertEqual(len(self.explorer.eth_getCode(self.contract_address)), 21678)
        self.assertEqual(len(self.explorer.eth_getBlockByHash(self.block_hash)), 20)
        self.assertEqual(len(self.explorer.eth_getBlockByNumber(self.block_number)), 20)

        self.assertEqual(len(self.explorer.eth_getTransactionByHash(self.tx_hash)), 14)
        self.assertEqual(len(self.explorer.eth_getTransactionByBlockNumberAndIndex(self.block_number, 1)), 14)

        self.assertEqual(len(self.explorer.eth_getTransactionReceipt('0xf02ffa405bae96e62a9e36fbd781362ca378ec62353d5e2bd0585868d3deaf61')), 12)
        # self.assertEqual(type(self.explorer.eth_newBlockFilter()), str)


if __name__ == '__main__':
    suite = unittest.TestLoader().loadTestsFromTestCase(EthereumExplorerTestCase)
    unittest.TextTestRunner(verbosity=2).run(suite)



class Instruction(object):
    """Instruction """
    def __init__(self, opcode, name,
                 operand_size, pops, pushes, fee,
                 description, operand=None,
                 operand_interpretation=None, offset=0, xref=None):

        self.opcode = opcode
        self.offset = offset
        self.name = name
        self.description = description
        self.operand_size = operand_size
        self.operand = operand  # Immediate operand if any
        # specific interpretation of operand value
        # ex: operand value for JUMP = xref
        self.operand_interpretation = operand_interpretation
        self.pops = pops
        self.pushes = pushes
        self.fee = fee
        self.xref = xref
        self.ssa = None

    def __eq__(self, other):
        """ Instructions are equal if all features match  """
        return self.opcode == other.opcode and\
            self.name == other.name and\
            self.operand == other.operand and\
            self.operand_size == other.operand_size and\
            self.pops == other.pops and\
            self.pushes == other.pushes and\
            self.fee == other.fee and\
            self.offset == other.offset and\
            self.description == other.description

    def __simple_output_format(self, offset=True):
        output = self.name
        if self.has_operand:
            output += ' 0x%x' % int.from_bytes(self.operand,
                                               byteorder='big')

        if offset:
            return "%d %s" % (self._offset, output)
        else:
            return "%s" % output

    #def __repr__(self):
    #    """ Entire representation of the instruction
    #    output = 'Instruction(0x%x, %r, %d, %d, %d, %d, %r, %r, %r)' \
    #        % (self._opcode, self._name, self._operand_size,
    #            self._pops, self._pushes, self._fee,
    #            self._description, self._operand, self._offset)"""
    #    return self.__simple_output_format()

    def __str__(self):
        """ String representation of the instruction """
        return self.__simple_output_format(offset=False)

    @property
    def bytes(self):
        """ Encoded instruction """
        byte = bytearray()
        byte.append(self.opcode)
        if self.operand:
            [byte.append(x) for x in self.operand]
        return "".join(map(chr, byte))

    @property
    def offset_end(self):
        """ Location in the program (optional) """
        return self.offset + self.size - 1

    @property
    def semantics(self):
        """ Canonical semantics """
        return self.name

    @property
    def size(self):
        """ Size of the encoded instruction """
        return self.operand_size + 1

    @property
    def has_operand(self):
        """ True if the instruction uses an immediate operand """
        return self.operand_size > 0

    @property
    def is_branch_conditional(self):
        """ Return list if the instruction is a jump """
        raise NotImplementedError

    @property
    def is_branch_unconditional(self):
        """ Return list if the instruction is a jump """
        raise NotImplementedError

    @property
    def is_branch(self):
        """ True if the instruction is a jump """
        return self.is_branch_conditional or self.is_branch_unconditional

    @property
    def is_halt(self):
        """ Return list if the instruction is a basic block terminator """
        raise NotImplementedError

    @property
    def is_terminator(self):
        """ True if the instruction is a basic block terminator """
        raise NotImplementedError

    @property
    def have_xref(self):
        """ TODO """
        raise NotImplementedError



import unittest

from octopus.arch.wasm.cfg import WasmCFG

EXAMPLE_PATH = "examples/wasm/samples/"


class WasmCFGraphTestCase(unittest.TestCase):

    def testCallGraph(self):
        def callgraph(bytecode, result_nodes, length_edges, fname=False):
            cfg = WasmCFG(bytecode)
            nodes, edges = cfg.get_functions_call_edges(fname)
            # visualize
            # cfg.visualize_call_flow()
            self.assertEqual(nodes, result_nodes)
            self.assertEqual(len(edges), length_edges)

        def controlflowgraph(bytecode, len_func, len_blocks, len_edges):
            cfg = WasmCFG(bytecode)
            functions = cfg.functions
            basicblocks = cfg.basicblocks
            edges = cfg.edges
            # visualize
            # cfg.visualize_call_flow()
            self.assertEqual(len(functions), len_func)
            self.assertEqual(len(basicblocks), len_blocks)
            self.assertEqual(len(edges), len_edges)

        def read_file(file_name):
            with open(file_name, 'rb') as f:
                module_bytecode = f.read()
            return module_bytecode

        # Helloworld
        module_bytecode = read_file(EXAMPLE_PATH + "helloworld.wasm")
        r_nodes = ['print(i32)', 'main()']
        r_nodes2 = ['print', 'main']
        callgraph(module_bytecode, r_nodes, 1, fname=True)
        callgraph(module_bytecode, r_nodes2, 1, fname=False)
        controlflowgraph(module_bytecode, 1, 1, 0)

        # fibonacci
        module_bytecode = read_file(EXAMPLE_PATH + "fib.wasm")
        r_nodes = ['i32 fib(i32)']
        r_nodes2 = ['fib']
        len_call_edges = 2
        callgraph(module_bytecode, r_nodes, len_call_edges, fname=True)
        callgraph(module_bytecode, r_nodes2, len_call_edges, fname=False)
        controlflowgraph(module_bytecode, 1, 3, 2)

        # fibonacci 2 with if-else
        module_bytecode = read_file(EXAMPLE_PATH + "fibc.wasm")
        r_nodes = ['i32 _fib(i32)', 'runPostSets()', '__post_instantiate()']
        r_nodes2 = ['_fib', 'runPostSets', '__post_instantiate']
        len_call_edges = 2
        callgraph(module_bytecode, r_nodes, len_call_edges, fname=True)
        callgraph(module_bytecode, r_nodes2, len_call_edges, fname=False)
        controlflowgraph(module_bytecode, 3, 10, 9)

if __name__ == '__main__':
    suite = unittest.TestLoader().loadTestsFromTestCase(WasmCFGraphTestCase)
    unittest.TextTestRunner(verbosity=2).run(suite)



from octopus.core.instruction import Instruction
from octopus.arch.wasm.wasm import _groups


class WasmInstruction(Instruction):
    """Wasm Instruction
    TODO

    """
    def __init__(self, opcode, name, imm_struct, operand_size, insn_byte,
                 pops, pushes, description, operand_interpretation=None, offset=0):
        """ TODO """
        self.opcode = opcode
        self.offset = offset
        self.name = name
        self.description = description
        self.operand_size = operand_size
        if len(insn_byte) > 1:
            self.operand = insn_byte[-operand_size:]  # Immediate operand if any
        else:
            self.operand = None
            # specific interpretation of operand value
        self.operand_interpretation = operand_interpretation
        self.insn_byte = insn_byte
        self.pops = pops
        self.pushes = pushes
        self.imm_struct = imm_struct
        self.xref = list()
        self.ssa = None

    def __eq__(self, other):
        """ Instructions are equal if all features match  """
        return self.opcode == other.opcode and\
            self.name == other.name and\
            self.insn_byte == other.insn_byte and\
            self.operand_size == other.operand_size and\
            self.pops == other.pops and\
            self.pushes == other.pushes and\
            self.operand_interpretation == other.operand_interpretation and\
            self.description == other.description

    def __str__(self):
        """ String representation of the instruction """
        if self.operand:
            return self.operand_interpretation
        # elif self.operand:
        #    return self.name + str(self.operand)
        else:
            return self.name

    @property
    def group(self):
        """ Instruction classification per group """
        last_class = _groups.get(0)
        for k, v in _groups.items():
            if self.opcode >= k:
                last_class = v
            else:
                return last_class
        return last_class

    @property
    def is_control(self):
        return self.group == 'Control'

    @property
    def is_parametric(self):
        return self.group == 'Parametric'

    @property
    def is_variable(self):
        return self.group == 'Variable'

    @property
    def is_memory(self):
        return self.group == 'Memory'

    @property
    def is_constant(self):
        return self.group == 'Constant'

    @property
    def is_logical_i32(self):
        return self.group == 'Logical_i32'

    @property
    def is_logical_i64(self):
        return self.group == 'Logical_i64'

    @property
    def is_logical_f32(self):
        return self.group == 'Logical_f32'

    @property
    def is_logical_f64(self):
        return self.group == 'Logical_f64'

    @property
    def is_arithmetic_i32(self):
        return self.group == 'Arithmetic_i32'

    @property
    def is_bitwise_i32(self):
        return self.group == 'Bitwise_i32'

    @property
    def is_arithmetic_i64(self):
        return self.group == 'Arithmetic_i64'

    @property
    def is_bitwise_i64(self):
        return self.group == 'Bitwise_i64'

    @property
    def is_arithmetic_f32(self):
        return self.group == 'Arithmetic_f32'

    @property
    def is_arithmetic_f64(self):
        return self.group == 'Arithmetic_f64'

    @property
    def is_conversion(self):
        return self.group == 'Conversion'

    @property
    def is_branch_conditional(self):
        """ Return True if the instruction is a conditional jump """
        return self.name in ['br_if', 'br_table', 'if']

    @property
    def is_branch_unconditional(self):
        """ Return True if the instruction is a unconditional jump """
        return self.name in ['br']

    @property
    def is_call(self):
        """ True if the instruction is a call instruction """
        return self.name in ['call', 'call_indirect']

    @property
    def is_branch(self):
        return self.is_branch_conditional or self.is_branch_unconditional

    @property
    def is_halt(self):
        """ Return True if the instruction is a branch terminator """
        return self.name in ['unreachable', 'return']

    @property
    def is_terminator(self):
        """ True if the instruction is a basic block terminator """
        return self.is_branch or self.is_halt

    @property
    def is_block_starter(self):
        """ Return True if the instruction is a basic block starter """
        return self.name in ['block', 'loop', 'if', 'else']

    @property
    def is_block_terminator(self):
        """ Return True if the instruction is a basic block terminator """
        return self.name in ['else', 'end']



from octopus.platforms.BTC.explorer import BitcoinExplorerRPC
from octopus.platforms.BTC.explorer import RPC_USER, RPC_PASSWORD, RPC_HOST

import unittest


class BitcoinExplorerTestCase(unittest.TestCase):

    explorer = BitcoinExplorerRPC(host=('%s:%s@%s' % (RPC_USER, RPC_PASSWORD, RPC_HOST)))

    blockhash = '00000000000000000024fb37364cbf81fd49cc2d51c09c75c35433c3a1945d04'
    txid = '1b5bfc2681d40c872126919ccb1752de4cca42dcfc594899f2ef11db4b05bb39'
    tx_raw = '0200000001686b654b40737f0daa1532f64e525dc925e60d075403d38cfb12ac9097764015040000006a473044022009ec3f26984906a813faae05d968ec06bf1c68883e09a00b6333126ea87d96b302201cf1d2b9165442aa178fdf772a3909c3d2ba69e454eb8660fa35df8645e3bcb60121022f2caec3ad2f3b174d048a0d46f4f6e8ba4e9d02f6bdbba64ac6817f7ac6c131ffffffff02060d0700000000001976a91407c5acae3abc91735a1471e275e33abbffada89088ac00581300000000001976a91432f2e30111e1dc45f415430ef082cb64225c538a88ac00000000'
    wallet_address = '15wDxrRCn7YiCXdvqjcih6G8svrmq5AQSS'
    script_hex = "76a82096b3fe1f4ec8fd076379267f72443bed81cc49c18a2913f7e1f0727f6f9f4fbf88ac"
    script_asm = 'OP_DUP OP_SHA256 96b3fe1f4ec8fd076379267f72443bed81cc49c18a2913f7e1f0727f6f9f4fbf OP_EQUALVERIFY OP_CHECKSIG'

    def testRPCCommand(self):

        #######################
        # HIGHT-LEVEL METHODS #
        #######################

        self.assertEqual(self.explorer.get_transaction(self.txid, 0), self.tx_raw)
        self.assertEqual(len(self.explorer.get_block_by_hash(self.blockhash)), 18)
        self.assertEqual(len(self.explorer.get_block_by_number(500000)), 18)

        ####################
        # JSON-RPC METHODS #
        ####################

        self.assertEqual(self.explorer.decoderawtransaction(self.tx_raw)['txid'], self.txid)
        self.assertEqual(self.explorer.decodescript(self.script_hex)['asm'], self.script_asm)
        self.assertEqual(len(self.explorer.getbestblockhash()), len(self.blockhash))
        self.assertEqual(len(self.explorer.getblock(self.blockhash)), 18)
        self.assertEqual(len(self.explorer.getblockchaininfo()), 11)
        self.assertEqual(type(self.explorer.getblockcount()), int)
        self.assertEqual(self.explorer.getblockhash(500000), self.blockhash)
        # self.assertEqual(len(self.explorer.getchaintips()), 2)
        self.assertEqual(type(self.explorer.getconnectioncount()), int)
        self.assertEqual(type(self.explorer.getdifficulty()), float)
        self.assertEqual(len(self.explorer.getinfo()), 16)
        self.assertEqual(len(self.explorer.getmempoolinfo()), 5)
        self.assertEqual(len(self.explorer.getmininginfo()), 8)
        self.assertEqual(len(self.explorer.getnettotals()), 4)
        self.assertEqual(type(self.explorer.getnetworkhashps()), float)
        self.assertEqual(len(self.explorer.getnetworkinfo()), 13)
        self.assertEqual(len(self.explorer.getpeerinfo()), 8)
        self.assertEqual(type(self.explorer.getrawmempool()), list)
        self.assertEqual(self.explorer.getrawtransaction(self.txid), self.tx_raw)
        self.assertEqual(type(self.explorer.getreceivedbyaccount('')), float)
        self.assertEqual(type(self.explorer.getreceivedbyaddress(self.wallet_address)), float)
        self.assertEqual(len(self.explorer.gettxout(self.txid, 0)), 5)
        self.assertEqual(len(self.explorer.gettxoutproof([self.txid])), 818)
        self.assertEqual(type(self.explorer.getunconfirmedbalance()), float)
        self.assertEqual(len(self.explorer.getwalletinfo()), 9)
        self.assertEqual(type(self.explorer.help()), str)
        self.assertEqual(len(self.explorer.validateaddress(self.wallet_address)), 6)
        self.assertEqual(self.explorer.verifytxoutproof(self.explorer.gettxoutproof([self.txid])), [self.txid])

        # Not tested
        '''
        self.explorer.abandontransaction()
        self.explorer.addmultisigaddress()
        self.explorer.addnode()
        self.explorer.createmultisig()
        self.explorer.createrawtransaction()
        self.explorer.dumpprivkey()
        self.explorer.encryptwallet()
        self.explorer.estimatefee()
        self.explorer.estimatepriority()
        self.explorer.getaccountaddress()
        self.explorer.getaccount()
        self.explorer.getaddednodeinfo()
        self.explorer.getaddressesbyaccount()
        self.explorer.getbalance()
        self.explorer.gettransaction()
        self.explorer.keypoolrefill()
        self.explorer.listaccounts()
        self.explorer.listaddressgroupings()
        self.explorer.listlockunspent()
        self.explorer.listreceivedbyaccount()
        self.explorer.listreceivedbyaddress()
        self.explorer.listtransactions()
        self.explorer.listunspent()
        self.explorer.lockunspent()
        self.explorer.prioritisetransaction()
        self.explorer.sendfrom()
        self.explorer.sendmany()
        self.explorer.sendrawtransaction()
        self.explorer.sendtoaddress()
        self.explorer.settxfee()
        self.explorer.signmessage()
        self.explorer.signrawtransaction()
        self.explorer.submitblock()

        self.explorer.verifymessage()
        self.explorer.walletlock()
        self.explorer.walletpassphrase()
        self.explorer.walletpassphrasechange()
        '''

if __name__ == '__main__':
    suite = unittest.TestLoader().loadTestsFromTestCase(BitcoinExplorerTestCase)
    unittest.TextTestRunner(verbosity=2).run(suite)



import io
import logging
import re

from octopus.engine.disassembler import Disassembler

from octopus.arch.evm.instruction import EvmInstruction
from octopus.arch.evm.evm import EVM


class EvmDisassembler(Disassembler):

    def __init__(self, bytecode=None):
        Disassembler.__init__(self, asm=EVM(), bytecode=bytecode)
        self.loader_code = None
        self.swarm_hash = None

    def runtime_code_detector(self):
        '''Check for presence of runtime code
        '''
        result = list(re.finditer('60.{2}604052', self.bytecode))
        if len(result) > 1:
            position = result[1].start()
            logging.info("[+] Runtime code detected")
            self.loader_code = self.bytecode[:position]
            self.bytecode = self.bytecode[position:]

    def swarm_hash_detector(self):
        '''Check for presence of Swarm hash at the end of bytecode
            https://github.com/ethereum/wiki/wiki/Swarm-Hash
        '''
        # we reduce to last 50 bytes to be sure it's the swarm hash
        # and not fake code
        swarm_hash_off = self.bytecode[-100:].find('a165627a7a72')
        # bzzr == 0x65627a7a72
        if swarm_hash_off > 0:
            logging.info("[+] Swarm hash detected in bytecodes")
            swarm_hash = self.bytecode[-100 + swarm_hash_off:]
            logging.info("[+] Swarm hash value: 0x%s", swarm_hash)
            logging.info("[+] Swarm hash removed")
            self.swarm_hash = self.bytecode[-100 + swarm_hash_off:]
            self.bytecode = self.bytecode[:-len(self.swarm_hash)]

    def analysis(self):
        self.runtime_code_detector()
        self.swarm_hash_detector()

    def disassemble_opcode(self, bytecode, offset=0):
        """
        TODO
        """

        wallet = io.BytesIO(bytecode)
        opcode = int.from_bytes(wallet.read(1), byteorder='big')

        # default value
        invalid = ('INVALID', 0, 0, 0, 0, 'Unknown opcode')
        name, operand_size, pops, pushes, gas, description = \
            self.asm.table.get(opcode, invalid)
        instruction = EvmInstruction(opcode, name, operand_size, pops, pushes,
                                     gas, description, offset=offset)
        if instruction.has_operand:
            instruction.operand = wallet.read(operand_size)
            if instruction.is_push:
                # directly calculate the operand int representation
                instruction.operand_interpretation = \
                    int.from_bytes(instruction.operand, byteorder='big')

        return instruction

    def disassemble(self, bytecode=None, offset=0, r_format='list',
                    analysis=True):
        '''
        creation code remove if analysis param is set to True (default)
        r_format: ('list' | 'text' | 'reverse')
        '''

        self.bytecode = bytecode if bytecode else self.bytecode

        if analysis:
            self.analysis()

        # reset lists
        self.instructions = list()
        self.reverse_instructions = dict()

        # call generic Disassembler.disassemble method
        return super().disassemble(self.bytecode, offset,
                                   r_format)



from octopus.platforms.BTC.disassembler import BitcoinDisassembler
import unittest


class BtcDisassemblerTestCase(unittest.TestCase):

    def testDisassembleHex(self):
        def test(bytecode_hex, result):
            disasm = BitcoinDisassembler(bytecode_hex).disassemble(r_format='text')
            self.assertEqual(disasm, result)

        def testNumberInstructions(bytecode_hex, len_result):
            disasm = BitcoinDisassembler(bytecode_hex).disassemble()
            self.assertEqual(len(disasm), len_result)

        # good test case : https://ivy-lang.org/bitcoin

        # https://en.bitcoin.it/wiki/Transaction#Pay-to-PubkeyHash
        # classic scriptPubKey
        bytecode_hex = "76a9149ba386253ea698158b6d34802bb9b550f5ce36dd88ac"
        result = 'OP_DUP\nOP_HASH160\n9ba386253ea698158b6d34802bb9b550f5ce36dd\nOP_EQUALVERIFY\nOP_CHECKSIG'
        test(bytecode_hex, result)
        testNumberInstructions(bytecode_hex, 5)

        # LockWithMultisig
        bytecode_hex = "007b7b522102f3ce98f0f47f9b91156b15cc43dd0956dfb1b94f2cacd95856ee1473a8c0ad6a2103399e680b0dc1f53c0d4641ef5f4c1267b69125b5076e1230f6ff1f4a1c4286eb21021cc83c35ab6cf47f6aa473449218e49dfa6e33ec2f0eb9e4e1ff709848ebe66453ae"
        result = '0\nOP_ROT\nOP_ROT\n2\n02f3ce98f0f47f9b91156b15cc43dd0956dfb1b94f2cacd95856ee1473a8c0ad6a\n03399e680b0dc1f53c0d4641ef5f4c1267b69125b5076e1230f6ff1f4a1c4286eb\n021cc83c35ab6cf47f6aa473449218e49dfa6e33ec2f0eb9e4e1ff709848ebe664\n3\nOP_CHECKMULTISIG'
        test(bytecode_hex, result)
        testNumberInstructions(bytecode_hex, 9)

        # LockWithPublicKeyHash
        bytecode_hex = "76a82096b3fe1f4ec8fd076379267f72443bed81cc49c18a2913f7e1f0727f6f9f4fbf88ac"
        result = 'OP_DUP\nOP_SHA256\n96b3fe1f4ec8fd076379267f72443bed81cc49c18a2913f7e1f0727f6f9f4fbf\nOP_EQUALVERIFY\nOP_CHECKSIG'
        test(bytecode_hex, result)
        testNumberInstructions(bytecode_hex, 5)

        # RevealPreImage
        bytecode_hex = 'a820c0c89d16be9b625834782a90411ebb259040775ee921651979d15fed5da0c26987'
        result = 'OP_SHA256\nc0c89d16be9b625834782a90411ebb259040775ee921651979d15fed5da0c269\nOP_EQUAL'
        test(bytecode_hex, result)
        testNumberInstructions(bytecode_hex, 3)

        # RevealCollision
        bytecode_hex = '6e879169a77ca787'
        result = 'OP_2DUP\nOP_EQUAL\nOP_NOT\nOP_VERIFY\nOP_SHA1\nOP_SWAP\nOP_SHA1\nOP_EQUAL'
        test(bytecode_hex, result)
        testNumberInstructions(bytecode_hex, 8)

        # RevealFixedPoint
        bytecode_hex = '76a887'
        result = 'OP_DUP\nOP_SHA256\nOP_EQUAL'
        test(bytecode_hex, result)
        testNumberInstructions(bytecode_hex, 3)

        bytecode_hex = '00b17521020cd5ac6e63926130ebff8b7063fa2cc26f597d6babed4adac23bff79e2178cceac'
        result = '0\nOP_CHECKLOCKTIMEVERIFY\nOP_DROP\n020cd5ac6e63926130ebff8b7063fa2cc26f597d6babed4adac23bff79e2178cce\nOP_CHECKSIG'
        test(bytecode_hex, result)
        testNumberInstructions(bytecode_hex, 5)

        # LockDelay
        bytecode_hex = '210361256041a4496060fd78bc5a3584622166210a6aa190513393541a47ccf8a67bad012ab27551'
        result = '0361256041a4496060fd78bc5a3584622166210a6aa190513393541a47ccf8a67b\nOP_CHECKSIGVERIFY\n42\nOP_CHECKSEQUENCEVERIFY\nOP_DROP\n1'
        test(bytecode_hex, result)
        testNumberInstructions(bytecode_hex, 6)

        # TransferWithTimeout
        bytecode_hex = '2102e5707a829f684e9c9263872ecd30c0027ca1a558b22f08f2aafc33f7b2eabdc27c630145b17567ad2103edb42eb19ef3badb27a5f7d13843dd8a58de5c0f6b9d9f1ffbeca59e9d35b85b68ac'
        result = '02e5707a829f684e9c9263872ecd30c0027ca1a558b22f08f2aafc33f7b2eabdc2\nOP_SWAP\nOP_IF\n69\nOP_CHECKLOCKTIMEVERIFY\nOP_DROP\nOP_ELSE\nOP_CHECKSIGVERIFY\n03edb42eb19ef3badb27a5f7d13843dd8a58de5c0f6b9d9f1ffbeca59e9d35b85b\nOP_ENDIF\nOP_CHECKSIG'
        test(bytecode_hex, result)
        testNumberInstructions(bytecode_hex, 11)

        # EscrowWithDelay
        bytecode_hex = '2103d32df19c6a163f2c0b3f047756e9fc4765c41897f59d5029ea624d839d6391b27c63ad0350c300b27551670072522102930524687b853a7bb80b5f24748e4b926869d629eca69e6e2d7758417e3249b02102d4e2e6a73d19d7eb52df220a826db2b825282d3ac700d75222813094a7583afc567a53ae68'
        result = '03d32df19c6a163f2c0b3f047756e9fc4765c41897f59d5029ea624d839d6391b2\nOP_SWAP\nOP_IF\nOP_CHECKSIGVERIFY\n50000\nOP_CHECKSEQUENCEVERIFY\nOP_DROP\n1\nOP_ELSE\n0\nOP_2SWAP\n2\n02930524687b853a7bb80b5f24748e4b926869d629eca69e6e2d7758417e3249b0\n02d4e2e6a73d19d7eb52df220a826db2b825282d3ac700d75222813094a7583afc\n6\nOP_ROLL\n3\nOP_CHECKMULTISIG\nOP_ENDIF'
        test(bytecode_hex, result)
        testNumberInstructions(bytecode_hex, 19)

        bytecode_hex = '6a4c50010002ca6efb0fe92dfa51e3393db37e0f4b79bda0634256a0158264dc2da9c5eda6262340d1510b03723e77839290a59735eb9f1dbb167caf1803887b3b92bbe824d800000000000000000000000000'
        result = 'OP_RETURN\n010002ca6efb0fe92dfa51e3393db37e0f4b79bda0634256a0158264dc2da9c5eda6262340d1510b03723e77839290a59735eb9f1dbb167caf1803887b3b92bbe824d800000000000000000000000000'
        test(bytecode_hex, result)
        testNumberInstructions(bytecode_hex, 2)

        # decimal value display and not hex as usual
        bytecode_hex = '6a03666666'
        result = 'OP_RETURN\n6710886'
        test(bytecode_hex, result)
        testNumberInstructions(bytecode_hex, 2)

        # FAIL but don't now why
        '''
        txid = '0415e0adb2ebb67fc48e53825eb8b86b3e32a45d863adae97b15de6dc4255cba'
        bytecode_hex = '6a03e29891'
        result = 'OP_RETURN -7235358'
        result_expected = 'OP_RETURN -1153250'
        test(bytecode_hex, result_expected)
        testNumberInstructions(bytecode_hex, 2)

        txid = '853eebe9937654e19956753d6f4f8c6ae5e4744b2e391756ce02bc5e638e108d'
        bytecode_hex = '6a52534b424c4f434b3a5841945ab918147d4e33f7e90dbebb68dd7dd1c9da273ca082cd66721943b9ff'
        result = 'OP_RETURN 2 3 424c4f434b3a5841945ab918147d4e33f7e90dbebb68dd7dd1c9da273ca082cd66721943b9ff'
        result_expected = 'OP_RETURN 2 3 [error]'
        test(bytecode_hex, result_expected)
        testNumberInstructions(bytecode_hex, 4)
        '''


if __name__ == '__main__':
    suite = unittest.TestLoader().loadTestsFromTestCase(BtcDisassemblerTestCase)
    unittest.TextTestRunner(verbosity=2).run(suite)


from octopus.engine.helper import helper as hlp
from z3 import UDiv, ULT, UGT

from logging import getLogger
logging = getLogger(__name__)

# =======================================
# #            SSA Simplifier           #
# =======================================


class EthereumSSASimplifier(object):
    def __init__(self):
        self._dispatch_function = {
            # arithmetic
            'ADD': self.symbolic_add,
            'SUB': self.operate_SUB,
            'MUL': self.operate_MUL,
            'DIV': self.operate_DIV,
            'MOD': self.operate_MOD,
            'SDIV': self.operate_SDIV,
            'SMOD': self.operate_SMOD,
            'ADDMOD': self.operate_ADDMOD,
            'MULMOD': self.operate_MULMOD,
            'EXP': self.operate_EXP,
            # SIGNEXTEND
            # logic
            'LT': self.operate_LT,
            'GT': self.operate_GT,
            'SLT': self.operate_SLT,
            'SGT': self.operate_SGT,
            'EQ': self.operate_EQ,
            'ISZERO': self.operate_ISZERO,
            'AND': self.operate_AND,
            'OR': self.operate_OR,
            'XOR': self.operate_XOR,
            # 'NOT': self.operate_NOT,
            # 'BYTE': self.operate_BYTE,
        }

    def is_all_constant(self, list):
        if [True] * len(list) == [i.ssa.is_constant for i in list]:
            return True
        return False

    def resolve_instr_ssa(self, instr):
        # SSA_TYPE_CONSTANT
        if instr.ssa.is_constant and instr.operand_interpretation:
            return instr.operand_interpretation
        # SSA_TYPE_FUNCTION
        elif instr.ssa.is_function:
            # args are all constants (most commun case)
            if self.is_all_constant(instr.ssa.args):
                values = [i.operand_interpretation for i in instr.ssa.args]
                result = self.symbolic_dispatcher(instr.name, values)
                # logging.warning('.. %x', result)
                return result
            else:
                # TODO: recursive resolve_instr_ir
                logging.warning('args not all constant')
                return None

        else:
            raise Exception('SSA_TYPE Unknown')
        return None

    def symbolic_dispatcher(self, mnemonic, values):

        fn = self._dispatch_function.get(mnemonic, None)
        return fn(*values)

    # arithmetics
    def symbolic_add(self, *values):
        s0 = hlp.convert_to_bitvec(values[0])
        s1 = hlp.convert_to_bitvec(values[1])
        return hlp.get_concrete_int(s0 + s1)

    def operate_SUB(self, *values):
        s0 = hlp.convert_to_bitvec(values[0])
        s1 = hlp.convert_to_bitvec(values[1])
        return hlp.get_concrete_int(s0 - s1)

    def operate_MUL(self, *values):
        s0 = hlp.convert_to_bitvec(values[0])
        s1 = hlp.convert_to_bitvec(values[1])
        return hlp.get_concrete_int(s0 * s1)

    def operate_DIV(self, *values):
        if values[1] == 0:
            return 0
        else:
            return hlp.get_concrete_int(UDiv(values[0] / values[1]))

    def operate_MOD(self, *values):
        return hlp.get_concrete_int(0 if values[1] == 0 else values[0] % values[1])

    def operate_SDIV(self, *values):
        s0, s1 = values[0], values[1]
        sign = -1 if (s0 / s1) < 0 else 1
        computed = sign * (abs(s0) / abs(s1))
        return hlp.get_concrete_int(computed)

    def operate_SMOD(self, *values):
        sign = -1 if values[0] < 0 else 1
        computed = sign * (abs(values[0]) % abs(values[1]))
        return hlp.get_concrete_int(computed)

    def operate_ADDMOD(self, *values):
        s0, s1, s2 = values[0], values[1], values[2]
        return hlp.get_concrete_int((s0 + s1) % s2 if s2 else 0)

    def operate_MULMOD(self, *values):
        s0, s1, s2 = values[0], values[1], values[2]
        return hlp.get_concrete_int((s0 * s1) % s2 if s2 else 0)

    def operate_EXP(self, *values):
        base, exponent = values[0], values[1]
        return hlp.get_concrete_int(pow(base, exponent))

        # logic
    def operate_LT(self, *values):
        s0 = hlp.convert_to_bitvec(values[0])
        s1 = hlp.convert_to_bitvec(values[1])
        return hlp.get_concrete_int(ULT(s0, s1))

    def operate_GT(self, *values):
        s0 = hlp.convert_to_bitvec(values[0])
        s1 = hlp.convert_to_bitvec(values[1])
        return hlp.get_concrete_int(UGT(s0, s1))

    def operate_SLT(self, *values):
        s0 = hlp.convert_to_bitvec(values[0])
        s1 = hlp.convert_to_bitvec(values[1])
        return hlp.get_concrete_int(s0 < s1)

    def operate_SGT(self, *values):
        s0 = hlp.convert_to_bitvec(values[0])
        s1 = hlp.convert_to_bitvec(values[1])
        return hlp.get_concrete_int(s0 > s1)

    def operate_EQ(self, *values):
        return hlp.get_concrete_int(values[0] == values[1])

    def operate_ISZERO(self, *values):
        return hlp.get_concrete_int(values[0] == 0)

    def operate_AND(self, *values):
        s0 = hlp.convert_to_bitvec(values[0])
        s1 = hlp.convert_to_bitvec(values[1])
        return hlp.get_concrete_int(s0 & s1)

    def operate_OR(self, *values):
        s0 = hlp.convert_to_bitvec(values[0])
        s1 = hlp.convert_to_bitvec(values[1])
        return hlp.get_concrete_int(s0 | s1)

    def operate_XOR(self, *values):
        s0 = hlp.convert_to_bitvec(values[0])
        s1 = hlp.convert_to_bitvec(values[1])
        return hlp.get_concrete_int(s0 ^ s1)

    # def operate_NOT(self, *values):
    #    return hlp.get_concrete_int(values[0] & values[1])

    # def operate_BYTE(self, *values):
    #    return



class BasicBlock(object):
    """
    """
    def __init__(self, start_offset=0x00, start_instr=None,
                 name='block_default_name'):
        self.start_offset = start_offset
        self.start_instr = start_instr
        self.name = name
        self.end_offset = start_offset
        self.end_instr = start_instr
        self.instructions = list()

        self.states = []
        self.function_name = "unknown"

    @property
    def size(self):
        return self.end_offset - self.start_offset

    def __str__(self):
        out = ''
        line = ''
        line = str(self.start_offset) + ': ' + str(self.name) + '\n'
        line += 'start_instr = ' + str(self.start_instr.name) + '\n'
        line += 'size = ' + str(self.size) + '\n'
        line += 'end_offset = ' + str(self.end_offset) + '\n'
        line += 'end_instr = ' + str(self.end_instr.name) + '\n'
        line += 'function_name = ' + str(self.function_name) + '\n'
        out += line + '\n\n'
        return out

    def instructions_details(self, format='hex'):
        out = ''
        line = ''
        for i in self.instructions:
            line = '%x: ' % i.offset
            if i.operand is not None and not i.xref:
                line += '%s' % str(i)
            elif isinstance(i.xref, list):
                line += '%s %s' % (i.name, i.xref)
            elif isinstance(i.xref, int):
                line += '%s %x' % (i.name, i.xref)
            else:
                line += i.name + ' '

            out += line + '\n'
        return out

    def instructions_ssa(self, format='hex'):
        out = ''
        line = ''
        for i in self.instructions:
            line = '%x: ' % i.offset
            if i.ssa:
                line += '' + i.ssa.format()
            else:
                line += '[NO_SSA] ' + i.name
            out += line + '\n'
        return out



import unittest

from octopus.arch.wasm.disassembler import WasmDisassembler

EXAMPLE_PATH = "examples/wasm/samples/"


class WasmDisassemblerTestCase(unittest.TestCase):

    def testDisassemble(self):
        def disassemble(bytecode, result):
            disasm = len(WasmDisassembler(bytecode).disassemble())
            self.assertEqual(disasm, result)

        def disasmOne(bytecode, result):
            disasm = str(WasmDisassembler(bytecode).disassemble_opcode(bytecode))
            self.assertEqual(disasm, result)

        def disasm(bytecode, result):
            text = WasmDisassembler(bytecode).disassemble(r_format='text')
            self.assertEqual(text, result)

        def disasmModule(bytecode, result_func, result_insn):
            funcs = WasmDisassembler().disassemble_module(bytecode)
            self.assertEqual(len(funcs), result_func)
            self.assertEqual(sum([len(i) for i in funcs]), result_insn)

        def read_file(file_name):
            with open(file_name, 'rb') as f:
                module_bytecode = f.read()
            return module_bytecode

        # Simple
        bytecode = bytearray([2, 127, 65, 24, 16, 28, 65, 0, 15, 11])
        result = 'block -1\ni32.const 24\ncall 28\ni32.const 0\nreturn\nend'

        disassemble(bytecode, 6)
        disasmOne(b'\x02\x7f', 'block -1')
        disasm(bytecode, result)

        # Helloworld
        bytecode_hex = "0061736d0100000001110460017f0060017e0060000060027e7e00021b0203656e76067072696e746e000103656e76067072696e7473000003030202030404017000000503010001071903066d656d6f7279020004696e69740002056170706c7900030a20020600411010010b17004120100120001000413010012001100041c00010010b0b3f050041040b04504000000041100b0d496e697420576f726c64210a000041200b0e48656c6c6f20576f726c643a20000041300b032d3e000041c0000b020a000029046e616d6504067072696e746e0100067072696e7473010004696e697400056170706c790201300131"
        disasmModule(bytecode_hex, 2, 14)

        # Helloworld - 2
        module_bytecode = read_file(EXAMPLE_PATH + "helloworld.wasm")
        disasmModule(module_bytecode, 1, 3)

        # fibonacci
        module_bytecode = read_file(EXAMPLE_PATH + "fib.wasm")
        disasmModule(module_bytecode, 1, 20)

if __name__ == '__main__':
    suite = unittest.TestLoader().loadTestsFromTestCase(WasmDisassemblerTestCase)
    unittest.TextTestRunner(verbosity=2).run(suite)



EDGE_UNCONDITIONAL = 'unconditional'
EDGE_CONDITIONAL_TRUE = 'conditional_true'
EDGE_CONDITIONAL_FALSE = 'conditional_false'
EDGE_FALLTHROUGH = 'fallthrough'
EDGE_CALL = 'call'


class Edge:

    def __init__(self, node_from, node_to, edge_type=EDGE_UNCONDITIONAL,
                 condition=None):

        self.node_from = node_from
        self.node_to = node_to
        self.type = edge_type
        self.condition = condition

    def __str__(self):
        return str(self.as_dict())

    def __eq__(self, other):
        return self.node_from == other.node_from and\
            self.node_to == other.node_to and\
            self.type == other.type and\
            self.condition == other.condition

    def __hash__(self):
        return hash(('from', self.node_from,
                     'to', self.node_to,
                     'type', self.type,
                     'condition', self.condition))

    def as_dict(self):
        return {'from': str(self.node_from), 'to': str(self.node_to),
                'type': self.type, 'condition': self.condition}



SSA_TYPE_FUNCTION = "function"
SSA_TYPE_CONSTANT = "constant"


class SSA(object):
    '''TODO'''

    def __init__(self, new_assignement=None, method_name=None, args=None, instr_type=SSA_TYPE_FUNCTION):
        """ TODO """
        self.new_assignement = new_assignement
        self.method_name = method_name
        self.args = args
        self.instr_type = instr_type

    def detail(self):
        out = ''
        out += 'new_assignement = ' + str(self.new_assignement) + '\n'
        out += 'method_name = ' + str(self.method_name) + '\n'
        out += 'args = ' + str(self.args) + '\n'
        out += 'instr_type = ' + str(self.instr_type) + '\n'
        out += '\n\n'
        return out

    def format(self):
        out = ''
        if self.is_constant:
            if self.new_assignement != None:
                out += '%{:02X}'.format(self.new_assignement)
                out += ' = '
            if self.args != None:
                out += '#0x%X' % self.args
            #if self.method_name:
            #    out += ' (%s)' % self.method_name
        elif self.is_function:
            if self.new_assignement != None:
                out += '%{:02X}'.format(self.new_assignement)
                out += ' = '
            if self.method_name:
                out += '%s(' % self.method_name
            if self.args != None:
                out += ', '.join('%{:02X}'.format(arg.ssa.new_assignement) for arg in self.args)
            out += ')'
        else:
            raise Exception('ssa_format no instr_type ')
        return out

    @property
    def is_constant(self):
        """ TODO """
        return self.instr_type == SSA_TYPE_CONSTANT

    @property
    def is_function(self):
        """ TODO """
        return self.instr_type == SSA_TYPE_FUNCTION



from octopus.arch.wasm.cfg import WasmCFG


# Eos smart contract == wasm module
class EosCFG(WasmCFG):
    def __init__(self, module_bytecode):
        WasmCFG.__init__(self,
                         module_bytecode=module_bytecode)

    def visualize_instrs_per_funcs(self, show=True, save=True,
                                   out_filename="eos_func_analytic.png",
                                   fontsize=8):
        super().visualize_instrs_per_funcs(show, save, out_filename, fontsize)
